"""
Testcases for the hardware command system.    
"""

from uppaal2jetracer.controller.hardware_command_system import HardwareCommandHandler, \
    HardwareCommandResultType, TurnCommand, SetSpeedCommand, PotentialCollisionCommand, \
    FreeAheadCommand, InvalidParamTypeError, InvalidParamsAmountError, \
    InvalidParamValueError


def test_hardware_command_handler_get_command_none():
    """
    Test that the get_command method returns None when the command does not exist.
    """
    handler = HardwareCommandHandler()

    assert handler.get_command("test") is None


def test_hardware_command_handler_has_command_false():
    """
    Test that the has_command method returns False when the command does not exist.
    """
    handler = HardwareCommandHandler()

    assert not handler.has_command("test")


def test_hardware_command_handler_get_command_all_commands_not_none():
    """
    Test that the get_command method returns the correct command object for all commands.
    """
    handler = HardwareCommandHandler()

    assert isinstance(handler.get_command("turn"), TurnCommand)
    assert isinstance(handler.get_command("set_speed"), SetSpeedCommand)
    assert isinstance(handler.get_command("potential_collision"), PotentialCollisionCommand)
    assert isinstance(handler.get_command("free_ahead"), FreeAheadCommand)


def test_hardware_command_handler_has_command_all_commands_true():
    """
    Test that the has_command method returns True for all commands.
    """
    handler = HardwareCommandHandler()

    assert handler.has_command("turn")
    assert handler.has_command("set_speed")
    assert handler.has_command("potential_collision")
    assert handler.has_command("free_ahead")


def test_free_ahead_command_success():
    """
    Test that the free_ahead command runs successfully.
    """
    handler = HardwareCommandHandler()

    command = handler.get_command("free_ahead")
    result = handler.run_command(command, [])
    assert result.result_type == HardwareCommandResultType.SUCCESS


def test_potential_collision_command_success():
    """
    Test that the potential_collision command runs successfully.
    """
    handler = HardwareCommandHandler()

    command = handler.get_command("potential_collision")
    result = handler.run_command(command, [])
    assert result.result_type == HardwareCommandResultType.SUCCESS


def test_set_speed_command_success():
    """
    Test that the set_speed command runs successfully.
    """
    handler = HardwareCommandHandler()

    command = handler.get_command("set_speed")
    result = handler.run_command(command, [0.5])
    assert result.result_type == HardwareCommandResultType.SUCCESS


def test_set_speed_command_invalid_params_failure():
    """
    Test that the set_speed command fails when the parameter is invalid.
    """
    handler = HardwareCommandHandler()

    command = handler.get_command("set_speed")
    try:
        handler.run_command(command, ["invalid param"])
        assert False
    except InvalidParamTypeError:
        assert True


def test_set_speed_command_no_params_failure():
    """
    Test that the set_speed command fails when no parameters are provided.
    """
    handler = HardwareCommandHandler()

    command = handler.get_command("set_speed")
    try:
        handler.run_command(command, [])
        assert False
    except InvalidParamsAmountError:
        assert True


def test_turn_command_success():
    """
    Test that the turn command runs successfully.
    """
    handler = HardwareCommandHandler()

    command = handler.get_command("turn")
    result = handler.run_command(command, [90])
    assert result.result_type == HardwareCommandResultType.SUCCESS


def test_turn_command_invalid_params_failure():
    """
    Test that the turn command fails when the parameter is invalid.
    """
    handler = HardwareCommandHandler()

    command = handler.get_command("turn")
    try:
        handler.run_command(command, ["invalid param"])
        assert False
    except InvalidParamTypeError:
        assert True


def test_turn_command_no_params_failure():
    """
    Test that the turn command fails when no parameters are provided.
    """
    handler = HardwareCommandHandler()

    command = handler.get_command("turn")
    try:
        handler.run_command(command, [])
        assert False
    except InvalidParamsAmountError:
        assert True

def test_free_ahead_command_too_many_params_failure():
    """
    Test that the free_ahead command fails when too many parameters are provided.
    """
    handler = HardwareCommandHandler()

    command = handler.get_command("free_ahead")
    try:
        handler.run_command(command, [1])
        assert False
    except InvalidParamsAmountError:
        assert True

def test_potential_collision_command_too_many_params_failure():
    """
    Test that the potential_collision command fails when too many parameters are provided.
    """
    handler = HardwareCommandHandler()

    command = handler.get_command("potential_collision")
    try:
        handler.run_command(command, [1])
        assert False
    except InvalidParamsAmountError:
        assert True


def test_set_speed_command_invalid_param_value_failure():
    """
    Test that the set_speed command fails when the parameter value is invalid.
    """
    handler = HardwareCommandHandler()

    command = handler.get_command("set_speed")
    try:
        handler.run_command(command, [2])
        assert False
    except InvalidParamValueError:
        assert True